import sys
from pathlib import Path
from unittest import TestCase

from mesido._darcy_weisbach import friction_factor
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestPipeDiameterSizingExample(TestCase):
    def test_half_network_gone(self):
        """
        This test is to check if the optimization behaves as expected under pipe class optimization.
        The test uses a symmetrical network with three demands in the middle that can be provided
        from a source both left and right. The optimal solution is that the optimizer only uses
        the left source and the associated left pipes.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - That expected pipes are removed
        - Check that the Q is under the max for the selected pipe class.
        - Check that head losses are as expected for the selected diameter
        - Check that head loss equals zero for removed pipes
        - Same for hydraulic power, no idea why it is outcommented

        Missing:
        - For now we use the hardcoded pipe classes to allow for variable maximum velocity over the
        pipe classes.

        """
        root_folder = str(Path(__file__).resolve().parent.parent)
        sys.path.insert(1, root_folder)

        import examples.pipe_diameter_sizing.src.example  # noqa: E402, I100
        from examples.pipe_diameter_sizing.src.example import (
            PipeDiameterSizingProblem,
        )  # noqa: E402, I100

        base_folder = (
            Path(examples.pipe_diameter_sizing.src.example.__file__).resolve().parent.parent
        )

        del root_folder
        sys.path.pop(1)

        problem = run_esdl_mesido_optimization(
            PipeDiameterSizingProblem,
            base_folder=base_folder,
            esdl_file_name="2a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        feasibility = problem.solver_stats["return_status"]
        self.assertTrue((feasibility == "Optimal"))

        parameters = problem.parameters(0)
        diameters = {p: parameters[f"{p}.diameter"] for p in problem.hot_pipes}
        results = problem.extract_results()

        # Check that half the network is removed, i.e. 4 pipes. Note that it
        # is equally possible for the left or right side of the network to be
        # removed.
        self.assertEqual(
            len([d for d in diameters.values() if d == 0.0]),
            4,
            "More/less than 4 pipes have been removed",
        )
        # Check that the correct/specific 4 pipes on the left or 4 on the right have been removed
        pipes_removed = ["Pipe_8592", "Pipe_2927", "Pipe_9a6f", "Pipe_a718"]
        pipes_remained = ["Pipe_96bc", "Pipe_51e4", "Pipe_6b39", "Pipe_f9b0"]
        self.assertTrue(
            all(
                (elem in [k for k, d in diameters.items() if (d == 0.0)] for elem in pipes_remained)
            )
            or all(
                elem in [k for k, d in diameters.items() if (d == 0.0)] for elem in pipes_removed
            ),
            "The incorrect 4 pipes have been removed",
        )

        for pipe in problem.energy_system_components.get("heat_pipe", []):
            neighbour = problem.has_related_pipe(pipe)
            if neighbour and pipe not in problem.hot_pipes:
                pipe = problem.cold_to_hot_pipe(pipe)
            given_pipe_classes = problem.pipe_classes(pipe)
            chosen_pc = [
                pc
                for pc in given_pipe_classes
                if round(results[f"{pipe}__hn_pipe_class_{pc.name}"][0]) == 1.0
            ][0]
            np.testing.assert_array_less(
                results[f"{pipe}.Q"],
                chosen_pc.maximum_velocity * np.pi * (chosen_pc.inner_diameter / 2.0) ** 2 + 1.0e-6,
            )

        for pipe in problem.energy_system_components.get("heat_pipe", []):
            if results[f"{pipe}__hn_diameter"] == 0.0:
                # TODO: At the moment it is so that a pipe which is not placed (diameter == 0.) can
                # have head loss since there is an equivalent solution where simultaniously the
                # is_disconnected variable is also true disabling the head_loss constraints.
                # np.testing.assert_allclose(results[f"{pipe}.dH"][1:], 0., atol=1.e-12)
                pass
            else:
                # TODO: there is a mismatch in maximum velocity, the linearization is done with the
                #  global setting instead of the pipe class specific one
                pc = problem.get_optimized_pipe_class(pipe)
                ff = friction_factor(
                    pc.maximum_velocity,
                    pc.inner_diameter,
                    2.0e-4,
                    parameters[f"{pipe}.temperature"],
                )
                c_v = parameters[f"{pipe}.length"] * ff / (2 * 9.81) / pc.inner_diameter
                dh_max = c_v * pc.maximum_velocity**2
                dh_manual = dh_max * results[f"{pipe}.Q"][1:] / pc.area / pc.maximum_velocity
                np.testing.assert_allclose(-dh_manual, results[f"{pipe}.dH"][1:], atol=1.0e-12)

        # Ensure that the removed pipes do not have predicted hydraulic power values
        hydraulic_power_sum = 0.0
        for pipe in diameters.keys():
            if pipe in pipes_removed:
                hydraulic_power_sum += sum(abs(results[f"{pipe}.Hydraulic_power"]))
        self.assertEqual(hydraulic_power_sum, 0.0, "Hydraulic power exists for a removed pipe")

        # Hydraulic power = delta pressure * Q = f(Q^3), where delta pressure = f(Q^2)
        # The linear approximation of the 3rd order function should overestimate the hydraulic
        # power when compared to the product of Q and the linear approximation of 2nd order
        # function (delta pressure).
        hydraulic_power_sum = 0.0
        hydraulic_power_post_process = 0.0
        for pipe in diameters.keys():
            if pipe in pipes_remained:
                hydraulic_power_sum += sum(abs(results[f"{pipe}.Hydraulic_power"]))
                hydraulic_power_post_process += sum(
                    abs(
                        results[f"{pipe}.dH"]
                        * parameters[f"{pipe}.rho"]
                        * 9.81
                        * results[f"{pipe}.HeatOut.Q"]
                    )
                )

        self.assertGreater(hydraulic_power_sum, hydraulic_power_post_process)
        demand_matching_test(problem, results)
        energy_conservation_test(problem, results)
        heat_to_discharge_test(problem, results)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestPipeDiameterSizingExample()
    a.test_half_network_gone()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from numpy import nan

from .source import Source


class GeothermalSource(Source):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_subtype = "geothermal"

        self.target_flow_rate = nan



from mesido.pycml import Variable

from ._non_storage_component import _NonStorageComponent


class CheckValve(_NonStorageComponent):
    """
    The check valve allows the fluid to flow in only one direction. This is done with constraints
    in the HeatMixin.
    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **self.merge_modifiers(
                dict(
                    Q=dict(min=0.0),
                ),
                modifiers,
            ),
        )

        self.component_type = "check_valve"

        self.add_variable(Variable, "dH", min=0.0)

        self.add_equation(self.dH - (self.HeatOut.H - self.HeatIn.H))

        self.add_equation((self.HeatIn.Heat - self.HeatOut.Heat) / self.Heat_nominal)



from .heat_source import HeatSource


class AirWaterHeatPump(HeatSource):
    """
    The air water heat pump component is used to model the source behaviour of air water heat pumps.
    For now, it is just a source, but in the future this can be expanded.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_subtype = "air_water_heat_pump"



from dataclasses import dataclass


@dataclass
class Asset:
    """
    Dataclass for an asset containing the information from the esdl objects.

    asset_type: string for the esdl asset type.
    id: str of the esdl id of the asset.
    in_ports: by default none, otherwise list of ports.
    out_ports: by default none, otherwise list of ports.
    attributes: dict of all the specified attributes of the esdl asset.
    global_properties: all the global properties specified in the esdl like the carriers.
    """

    asset_type: str
    id: str
    name: str
    in_ports: None
    out_ports: None
    attributes: dict
    global_properties: dict



from __future__ import annotations

import datetime
import logging

import numpy as np

from rtctools.data.storage import DataStore


logger = logging.getLogger("WarmingUP-MPC")
logger.setLevel(logging.INFO)


def set_data_with_averages_and_peak_day(
    datastore: DataStore,
    variable_name: str,
    ensemble_member: int,
    new_date_times: np.array,
    problem: object,
):
    try:
        data = problem.get_timeseries(variable=variable_name, ensemble_member=ensemble_member)
    except KeyError:
        datastore.set_timeseries(
            variable=variable_name,
            datetimes=new_date_times,
            values=np.asarray([0.0] * len(new_date_times)),
            ensemble_member=ensemble_member,
            check_duplicates=True,
        )
        return

    new_data = list()
    data_timestamps = data.times
    data_datetimes = [
        problem.io.datetimes[0] + datetime.timedelta(seconds=s) for s in data_timestamps
    ]
    assert new_date_times[0] == data_datetimes[0]
    data_values = data.values

    values_for_mean = [0.0]
    for dt, val in zip(data_datetimes, data_values):
        if dt in new_date_times:
            new_data.append(np.mean(values_for_mean))
            values_for_mean = [val]
        else:
            values_for_mean.append(val)

    # last datetime is not in input data, so we need to take the mean of the last bit
    new_data.append(np.mean(values_for_mean))

    datastore.set_timeseries(
        variable=variable_name,
        datetimes=new_date_times,
        values=np.asarray(new_data),
        ensemble_member=ensemble_member,
        check_duplicates=True,
    )


def adapt_hourly_year_profile_to_day_averaged_with_hourly_peak_day(problem, problem_day_steps: int):
    """
    Adapt yearly porifle with hourly time steps to a common profile (daily averaged profile except
    for the day with the peak demand).

    Return the following:
        - problem_indx_max_peak: index of the maximum of the peak values
        - heat_demand_nominal: max demand value found for a specific heating demand
    """

    demands = problem.energy_system_components.get("heat_demand", [])
    new_datastore = DataStore(problem)
    new_datastore.reference_datetime = problem.io.datetimes[0]

    for ensemble_member in range(problem.ensemble_size):
        parameters = problem.parameters(ensemble_member)
        total_demand = None
        heat_demand_nominal = dict()

        for demand in demands:
            try:
                demand_values = problem.get_timeseries(
                    f"{demand}.target_heat_demand", ensemble_member
                ).values
            except KeyError:
                continue
            if total_demand is None:
                total_demand = demand_values
            else:
                total_demand += demand_values
            heat_demand_nominal[f"{demand}.Heat_demand"] = max(demand_values)
            heat_demand_nominal[f"{demand}.Heat_flow"] = max(demand_values)

        # TODO: the approach of picking one peak day was introduced for a network with a tree
        #  layout and all big sources situated at the root of the tree. It is not guaranteed
        #  that an optimal solution is reached in different network topologies.
        idx_max = int(np.argmax(total_demand))
        max_day = idx_max // 24
        nr_of_days = len(total_demand) // 24
        new_date_times = list()
        day_steps = problem_day_steps

        problem_indx_max_peak = max_day // day_steps
        if max_day % day_steps > 0:
            problem_indx_max_peak += 1.0

        for day in range(0, nr_of_days, day_steps):
            if day == max_day // day_steps * day_steps:
                if max_day > day:
                    new_date_times.append(problem.io.datetimes[day * 24])
                new_date_times.extend(problem.io.datetimes[max_day * 24 : max_day * 24 + 24])
                if (day + day_steps - 1) > max_day:
                    new_date_times.append(problem.io.datetimes[max_day * 24 + 24])
            else:
                new_date_times.append(problem.io.datetimes[day * 24])
        new_date_times.append(problem.io.datetimes[-1] + datetime.timedelta(hours=1))

        new_date_times = np.asarray(new_date_times)
        parameters["times"] = [x.timestamp() for x in new_date_times]

        for demand in demands:
            var_name = f"{demand}.target_heat_demand"
            set_data_with_averages_and_peak_day(
                datastore=new_datastore,
                variable_name=var_name,
                ensemble_member=ensemble_member,
                new_date_times=new_date_times,
                problem=problem,
            )

        # TODO: this has not been tested but is required if a production profile is included
        #  in the data
        for source in problem.energy_system_components.get("heat_source", []):
            var_name = f"{source}.maximum_heat_source"
            try:
                problem.get_timeseries(variable=var_name, ensemble_member=ensemble_member)
            except KeyError:
                logger.debug(
                    f"Source {source} has no production profile, thus it also will "
                    f"not be adapted to a different time scales."
                )
                continue

            set_data_with_averages_and_peak_day(
                datastore=new_datastore,
                variable_name=var_name,
                ensemble_member=ensemble_member,
                new_date_times=new_date_times,
                problem=problem,
            )

        for carrier_properties in problem.esdl_carriers.values():
            carrier_name = carrier_properties["name"]
            var_name = f"{carrier_name}.price_profile"
            try:
                problem.get_timeseries(variable=var_name, ensemble_member=ensemble_member)
            except KeyError:
                logger.debug(
                    f"Carrier {carrier_name} has no price profile, thus it also will "
                    f"not be adapted to different time scales."
                )
            set_data_with_averages_and_peak_day(
                datastore=new_datastore,
                variable_name=var_name,
                ensemble_member=ensemble_member,
                new_date_times=new_date_times,
                problem=problem,
            )

    problem.io = new_datastore

    logger.info("Profile data has been adapted to a common format")

    return problem_indx_max_peak, heat_demand_nominal



from mesido.pycml import Variable

from numpy import nan, pi

from .gas_base import GasTwoPort
from .._internal import BaseAsset


class GasPipe(GasTwoPort, BaseAsset):
    """
    The gas_pipe component is used to model head loss through the pipe. At the moment we only have
    a placeholder linear head loss formulation in place.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "gas_pipe"
        self.disconnectable = False

        self.v_max = 15.0
        self.density = 2.5e3  # [g/m3]
        self.rho = self.density
        self.diameter = nan
        self.area = 0.25 * pi * self.diameter**2
        self.Q_nominal = self.v_max / 2.0 * self.area
        self.pressure = 16.0e5
        self.id_mapping_carrier = -1

        self.nominal_head = 30.0
        self.length = nan
        self.r = 1.0e-6 * self.length  # TODO: temporary value
        self.nominal_head_loss = (self.Q_nominal * self.r * self.nominal_head) ** 0.5

        self.add_variable(Variable, "dH", nominal=self.Q_nominal * self.r)
        self.add_variable(Variable, "Q", nominal=self.Q_nominal)

        # Flow should be preserved
        self.add_equation(((self.GasIn.Q - self.GasOut.Q) / self.Q_nominal))
        self.add_equation(((self.Q - self.GasOut.Q) / self.Q_nominal))
        self.add_equation(((self.GasIn.Q - self.GasIn.mass_flow / self.density) / self.Q_nominal))

        # Flow should be preserved
        self.add_equation(
            ((self.GasIn.mass_flow - self.GasOut.mass_flow) / (self.Q_nominal * self.density))
        )
        # # shadow Q for aliases
        self.add_equation(((self.GasOut.Q_shadow - (self.GasIn.Q_shadow - 1.0e-3))))

        # Hydraulic power
        # TODO replace value
        # rho * ff * length * area / 2 / diameter * velocity**3
        ff = 0.02  # Order of magnitude expected with 0.05-2.5m/s in 20mm-1200mm diameter pipe
        velo = self.Q_nominal / self.area
        self.Hydraulic_power_nominal = (
            self.rho * ff * max(self.length, 1.0) * pi * self.area / self.diameter / 2.0 * velo**3
        )
        self.add_variable(
            Variable, "Hydraulic_power", min=0.0, nominal=self.Hydraulic_power_nominal
        )  # [W]

        self.add_equation(
            (self.Hydraulic_power - (self.GasIn.Hydraulic_power - self.GasOut.Hydraulic_power))
            / (self.pressure * self.Q_nominal * self.Hydraulic_power_nominal) ** 0.5
        )



"""All test for a electric node/bus

Currently both the MILP and NLP tests

What at least was implement
- voltages on all connections are the same
- power in == power out
"""

from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import electric_power_conservation_test


class TestMILPbus(TestCase):
    def test_voltages_and_power_network1(self):
        """
        Checks the behaviour of electricity networks with a bus asset. A bus asset is the
        only asset that is allowed to have more than one electricity port.

        Checks:
        - Voltage is equal for the bus ports
        - Checks energy conservation in the bus
        - Checks current conservation in the bus
        - Checks that minimum voltage is met
        - Checks that power meets the current * voltage at the demands

        """
        import models.unit_cases_electricity.bus_networks.src.example as example
        from models.unit_cases_electricity.bus_networks.src.example import ElectricityProblem

        base_folder = Path(example.__file__).resolve().parent.parent

        # Run the problem
        solution = run_esdl_mesido_optimization(
            ElectricityProblem,
            base_folder=base_folder,
            esdl_file_name="Electric_bus3.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        results = solution.extract_results()

        # electric power conservation system and no dissipation of power and current in bus
        electric_power_conservation_test(solution, results)

        v1 = results["Bus_f262.ElectricityConn[1].V"]
        v2 = results["Bus_f262.ElectricityConn[2].V"]
        v_outgoing_cable = results["ElectricityCable_de9a.ElectricityIn.V"]
        v_incoming_cable = results["ElectricityCable_1ad0.ElectricityOut.V"]
        v_demand = results["ElectricityDemand_e527.ElectricityIn.V"]
        p_demand = results["ElectricityDemand_e527.ElectricityIn.Power"]
        i_demand = results["ElectricityDemand_e527.ElectricityIn.I"]

        # Incoming voltage == outgoing voltage of bus
        self.assertTrue(all(v1 == v2))
        # Ingoing voltage of bus == voltage of incoming cable
        self.assertTrue(all(v1 == v_incoming_cable))
        # Outgoing voltage of bus == voltage of outgoing cable
        self.assertTrue(all(v1 == v_outgoing_cable))

        # check if minimum voltage is reached
        np.testing.assert_array_less(
            solution.parameters(0)["ElectricityDemand_e527.min_voltage"] - 1.0e-3, v_demand
        )
        # Check that current is high enough to carry the power
        np.testing.assert_array_less(p_demand - 1e-12, v_demand * i_demand)



import logging
from enum import IntEnum
from math import isclose
from typing import List, Tuple

import casadi as ca

from mesido.base_component_type_mixin import BaseComponentTypeMixin

import numpy as np

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)
from rtctools.optimization.timeseries import Timeseries


logger = logging.getLogger("mesido")


class ElectrolyzerOption(IntEnum):
    """
    Enumeration for the possible options to model the electrolyzer.
    For detailed calculations related to these options, see the equations themselves.

    CONSTANT_EFFICIENCY
        A constant efficiency is used to determine the H2 produced per energy electricity entering

    LINEARIZED_THREE_LINES_WEAK_INEQUALITY
        The efficiency curve is linearized in 3 lines, which are all inequalities and with proper
        goals will move towards these lines

    LINEARIZED_THREE_LINES_EQUALITY
        The efficiency curve is linearized in 3 lines, which are all equalities using binary
        variables and the big-M method to select the relevant lines.
    """

    CONSTANT_EFFICIENCY = 1
    LINEARIZED_THREE_LINES_WEAK_INEQUALITY = 2
    LINEARIZED_THREE_LINES_EQUALITY = 3


class ElectricityPhysicsMixin(BaseComponentTypeMixin, CollocatedIntegratedOptimizationProblem):
    """
    This class is used to model the physics of an electricity network with its assets. We model
    the different components with variety of linearization strategies.
    """

    def __init__(self, *args, **kwargs):
        """
        In this __init__ we prepare the dicts for the variables added by the HeatMixin class
        """

        super().__init__(*args, **kwargs)

        # Variable for when in time an asset switched on due to meeting a requirement
        self.__asset_is_switched_on_map = {}
        self.__asset_is_switched_on_var = {}
        self.__asset_is_switched_on_bounds = {}

        self.__electricity_producer_upper_bounds = {}

        self._electricity_cable_topo_cable_class_map = {}

        # Boolean path-variable for the charging of storage assets
        self.__storage_charging_var = {}
        self.__storage_charging_bounds = {}
        self.__storage_charging_map = {}

        self.__set_point_var = {}
        self.__set_point_bounds = {}
        self.__set_point_map = {}

        self.__electricity_storage_discharge_var = {}
        self.__electricity_storage_discharge_bounds = {}
        self.__electricity_storage_discharge_nominals = {}
        self.__electricity_storage_discharge_map = {}

    def energy_system_options(self):
        r"""
        Returns a dictionary of milp network specific options.
        """

        options = {}

        options["include_asset_is_switched_on"] = False
        options["include_electric_cable_power_loss"] = False
        options["electrolyzer_efficiency"] = (
            ElectrolyzerOption.LINEARIZED_THREE_LINES_WEAK_INEQUALITY
        )
        options["electricity_storage_discharge_variables"] = False

        return options

    def electricity_carriers(self):
        """
        This function should be overwritten by the problem and should give a dict with the
        carriers as keys and a list of voltage values.
        """
        return {}

    def pre(self):
        """
        In this pre method we fill the dicts initiated in the __init__. This means that we create
        the Casadi variables and determine the bounds, nominals and create maps for easier
        retrieving of the variables.
        """
        super().pre()

        options = self.energy_system_options()

        self.__update_electricity_producer_upper_bounds()

        if options["include_asset_is_switched_on"]:
            for asset in [
                *self.energy_system_components.get("electrolyzer", []),
            ]:
                var_name = f"{asset}__asset_is_switched_on"
                self.__asset_is_switched_on_map[asset] = var_name
                self.__asset_is_switched_on_var[var_name] = ca.MX.sym(var_name)
                self.__asset_is_switched_on_bounds[var_name] = (0.0, 1.0)

        for asset in [*self.energy_system_components.get("electricity_storage", [])]:
            var_name = f"{asset}__is_charging"
            self.__storage_charging_map[asset] = var_name
            self.__storage_charging_var[var_name] = ca.MX.sym(var_name)
            self.__storage_charging_bounds[var_name] = (0.0, 1.0)

            if options["electricity_storage_discharge_variables"]:
                bound_storage = self.bounds()[f"{asset}.Effective_power_charging"][0]
                var_name = f"{asset}__effective_power_discharging"
                self.__electricity_storage_discharge_map[asset] = var_name
                self.__electricity_storage_discharge_var[var_name] = ca.MX.sym(var_name)
                self.__electricity_storage_discharge_bounds[var_name] = (0, -bound_storage)
                self.__electricity_storage_discharge_nominals[var_name] = self.variable_nominal(
                    f"{asset}.Effective_power_charging"
                )

        for asset in [*self.energy_system_components.get("electricity_source", [])]:
            if isinstance(self.bounds()[f"{asset}.Electricity_source"][1], Timeseries):
                var_name = f"{asset}__set_point"
                self.__set_point_map[asset] = var_name
                self.__set_point_var[var_name] = ca.MX.sym(var_name)
                self.__set_point_bounds[var_name] = (0.0, 1.0)

    @property
    def extra_variables(self):
        """
        In this function we add all the variables defined in the HeatMixin to the optimization
        problem. Note that these are only the normal variables not path variables.
        """
        variables = super().extra_variables.copy()

        return variables

    @property
    def path_variables(self):
        """
        In this function we add all the path variables defined in the HeatMixin to the
        optimization problem. Note that path_variables are variables that are created for each
        time-step.
        """
        variables = super().path_variables.copy()

        variables.extend(self.__asset_is_switched_on_var.values())
        variables.extend(self.__storage_charging_var.values())
        variables.extend(self.__set_point_var.values())
        variables.extend(self.__electricity_storage_discharge_var.values())

        return variables

    def variable_is_discrete(self, variable):
        """
        All variables that only can take integer values should be added to this function.
        """

        if variable in self.__asset_is_switched_on_var:
            return True
        if variable in self.__storage_charging_var:
            return True
        else:
            return super().variable_is_discrete(variable)

    def variable_nominal(self, variable):
        """
        In this function we add all the nominals for the variables defined/added in the HeatMixin.
        """
        if variable in self.__electricity_storage_discharge_nominals:
            return self.__electricity_storage_discharge_nominals[variable]

        return super().variable_nominal(variable)

    def bounds(self):
        """
        In this function we add the bounds to the problem for all the variables defined/added in
        the HeatMixin.
        """
        bounds = super().bounds()

        bounds.update(self.__asset_is_switched_on_bounds)
        bounds.update(self.__storage_charging_bounds)
        bounds.update(self.__electricity_producer_upper_bounds)
        bounds.update(self.__set_point_bounds)
        bounds.update(self.__electricity_storage_discharge_bounds)

        return bounds

    @staticmethod
    def __get_abs_max_bounds(*bounds):
        """
        This function returns the absolute maximum of the bounds given. Note that bounds can also be
        a timeseries.
        """
        max_ = 0.0

        for b in bounds:
            if isinstance(b, np.ndarray):
                max_ = max(max_, max(abs(b)))
            elif isinstance(b, Timeseries):
                max_ = max(max_, max(abs(b.values)))
            else:
                max_ = max(max_, abs(b))

        return max_

    def __state_vector_scaled(self, variable, ensemble_member):
        """
        This functions returns the casadi symbols scaled with their nominal for the entire time
        horizon.
        """
        canonical, sign = self.alias_relation.canonical_signed(variable)
        return (
            self.state_vector(canonical, ensemble_member) * self.variable_nominal(canonical) * sign
        )

    def __update_electricity_producer_upper_bounds(self):
        t = self.times()

        timeseries_io_names = self.io.get_timeseries_names()
        for asset in self.energy_system_components.get("electricity_source", []):
            if f"{asset}.maximum_electricity_source" in timeseries_io_names:
                lb = Timeseries(t, np.zeros(len(t)))
                ub = self.get_timeseries(f"{asset}.maximum_electricity_source")
                self.__electricity_producer_upper_bounds[f"{asset}.Electricity_source"] = (lb, ub)

    def __electricity_producer_set_point_constraints(self, ensemble_member):
        """
        This function adds constraints for wind parks which generates electrical power. The
        produced electrical power is capped with a user specified percentage value of the maximum
        value.
        """
        constraints = []

        for asset in [*self.energy_system_components.get("electricity_source", [])]:
            if asset in self.__set_point_map.keys():
                var_name = self.__set_point_map[asset]
                set_point = self.__state_vector_scaled(var_name, ensemble_member)
                electricity_source = self.__state_vector_scaled(
                    f"{asset}.Electricity_source", ensemble_member
                )
                # TODO: [: len(self.times())] should be removed once the emerge test is properly
                # time-sampled.
                max = self.bounds()[f"{asset}.Electricity_source"][1].values[: len(self.times())]
                nominal = (
                    self.variable_nominal(f"{asset}.Electricity_source") * np.median(max)
                ) ** 0.5

                constraints.append(((set_point * max - electricity_source) / nominal, 0.0, 0.0))

        return constraints

    def __electricity_node_mixing_path_constraints(self, ensemble_member):
        """
        This function adds constraints for power/energy and current conservation at nodes/busses.
        """
        constraints = []

        for bus, connected_cables in self.energy_system_topology.busses.items():
            power_sum = 0.0
            i_sum = 0.0
            power_nominal = []
            i_nominal = []

            for i_conn, (_cable, orientation) in connected_cables.items():
                power_con = f"{bus}.ElectricityConn[{i_conn + 1}].Power"
                i_port = f"{bus}.ElectricityConn[{i_conn + 1}].I"
                power_sum += orientation * self.state(power_con)
                i_sum += orientation * self.state(i_port)
                power_nominal.append(self.variable_nominal(power_con))
                i_nominal.append(self.variable_nominal(i_port))

            power_nominal = np.median(power_nominal)
            constraints.append((power_sum / power_nominal, 0.0, 0.0))

            i_nominal = np.median(i_nominal)
            constraints.append((i_sum / i_nominal, 0.0, 0.0))

        return constraints

    def __electricity_cable_mixing_path_constraints(self, ensemble_member):
        """
        This function adds constraints relating the electrical power to the current flowing through
        the cable. The power through the cable is limited by the maximum voltage and the actual
        current variable with an inequality constraint. This is done to allow power losses through
        the network. As the current and power are related with an equality constraint at the
        demands exactly matching the P = U*I equation, we allow the inequalities for the lines. By
        overestimating the power losses and voltage drops, together we ensure that U*I>P.


        Furthermore, the power loss is estimated by linearizing with the maximum current, meaning
        that we are always overestimating the power loss in the cable.
        """
        constraints = []
        parameters = self.parameters(ensemble_member)

        for cable in self.energy_system_components.get("electricity_cable", []):
            current = self.state(f"{cable}.ElectricityIn.I")
            power_in = self.state(f"{cable}.ElectricityIn.Power")
            power_out = self.state(f"{cable}.ElectricityOut.Power")
            power_loss = self.state(f"{cable}.Power_loss")
            # v_loss = self.state(f"{cable}.V_loss")
            r = parameters[f"{cable}.r"]
            i_max = parameters[f"{cable}.max_current"]
            v_nom = parameters[f"{cable}.nominal_voltage"]
            v_max = parameters[f"{cable}.max_voltage"]
            # v_loss_nom = parameters[f"{cable}.nominal_voltage_loss"]
            length = parameters[f"{cable}.length"]

            # Ensure that the current is sufficient to transport the power
            constraints.append(((power_in - current * v_max) / (i_max * v_max), -np.inf, 0.0))
            constraints.append(((power_out - current * v_max) / (i_max * v_max), -np.inf, 0.0))
            # Power loss constraint
            options = self.energy_system_options()
            if options["include_electric_cable_power_loss"]:
                if cable in self._electricity_cable_topo_cable_class_map.keys():
                    cable_classes = self._electricity_cable_topo_cable_class_map[cable]
                    max_res = max([cc.resistance for cc in cable_classes])
                    max_i_max = max([cc.maximum_current for cc in cable_classes])
                    big_m = max_i_max**2 * max_res * length
                    constraint_nominal = max_i_max * v_nom * max_res * length
                    for cc_data, cc_name in cable_classes.items():
                        if cc_name != "None":
                            i_max = cc_data.maximum_current
                            res = cc_data.resistance
                            exp = current * res * length * i_max
                            is_selected = self.variable(cc_name)
                            constraints.append(
                                (
                                    (power_loss - exp + big_m * (1 - is_selected))
                                    / constraint_nominal,
                                    0.0,
                                    np.inf,
                                )
                            )
                            constraints.append(
                                (
                                    (power_loss - exp - big_m * (1 - is_selected))
                                    / (constraint_nominal),
                                    -np.inf,
                                    0.0,
                                )
                            )
                else:
                    constraints.append(
                        ((power_loss - current * r * i_max) / (i_max * v_nom * r), 0.0, 0.0)
                    )
            else:
                constraints.append(((power_loss) / (i_max * v_nom * r), 0.0, 0.0))

        return constraints

    def __voltage_loss_path_constraints(self, ensemble_member):
        """
        Furthermore, the voltage_loss symbol is set, as it depends on the chosen pipe
        class, e.g. the related resistance and the current through the cable.

        Parameters
        ----------
        ensemble_member : The ensemble of the optimization

        Returns
        -------
        list of the added constraints
        """
        constraints = []
        parameters = self.parameters(ensemble_member)

        for cable in self.energy_system_components.get("electricity_cable", []):
            cable_classes = []

            current = self.state(f"{cable}.ElectricityIn.I")
            v_loss = self.state(f"{cable}.V_loss")
            r = parameters[f"{cable}.r"]
            # v_loss_nom = parameters[f"{cable}.nominal_voltage_loss"]
            v_nom = parameters[f"{cable}.nominal_voltage"]
            c_length = parameters[f"{cable}.length"]

            constraint_nominal = self.variable_nominal(v_loss)

            # TODO: still have to check for proper scaling
            if cable in self._electricity_cable_topo_cable_class_map.keys():
                cable_classes = self._electricity_cable_topo_cable_class_map[cable]
                variables = {
                    cc.name: self.variable(var_name) for cc, var_name in cable_classes.items()
                }
                resistances = {cc.name: cc.resistance for cc in cable_classes}

                # to be updated for a better value, but it should also cover the gap between two
                # nodes when no cable is placed, so should be able to reach v_max
                big_m = v_nom

                for var_size, variable in variables.items():
                    if var_size != "None":
                        expr = resistances[var_size] * c_length * current
                        constraints.append(
                            (
                                (v_loss - expr + big_m * (1 - variable)) / constraint_nominal,
                                0.0,
                                np.inf,
                            )
                        )
                        constraints.append(
                            (
                                (v_loss - expr - big_m * (1 - variable)) / constraint_nominal,
                                -np.inf,
                                0.0,
                            )
                        )

            else:
                constraints.append(((v_loss - r * current) / constraint_nominal, 0.0, 0.0))

        return constraints

    def __electricity_demand_path_constraints(self, ensemble_member):
        """
        This function adds the constraints for the electricity commodity at the demand assets. We
        enforce that a minimum voltage is exactly met together with the power that is carried by
        the current. By fixing the voltage at the demand we ensure that at the demands
        P = U * I is met exactly at this point in the network and the power is conservatively
        in the cables at all locations in the network.
        """
        constraints = []
        parameters = self.parameters(ensemble_member)

        for elec_demand in [
            *self.energy_system_components.get("electricity_demand", []),
            *self.energy_system_components.get("heat_pump_elec", []),
            *self.energy_system_components.get("electrolyzer", []),
            *self.energy_system_components.get("transformer", []),
        ]:
            min_voltage = parameters[f"{elec_demand}.min_voltage"]
            voltage = self.state(f"{elec_demand}.ElectricityIn.V")
            # to ensure that voltage entering is equal or larger than the minimum voltage
            constraints.append(((voltage - min_voltage) / min_voltage, 0.0, np.inf))

            power_nom = self.variable_nominal(f"{elec_demand}.ElectricityIn.Power")
            curr_nom = self.variable_nominal(f"{elec_demand}.ElectricityIn.I")
            power_in = self.state(f"{elec_demand}.ElectricityIn.Power")
            current_in = self.state(f"{elec_demand}.ElectricityIn.I")

            constraints.append(
                (
                    (power_in - min_voltage * current_in)
                    / (power_nom * curr_nom * min_voltage) ** 0.5,
                    0,
                    0,
                )
            )

        return constraints

    def __electricity_storage_path_constraints(self, ensemble_member):
        """
        This function adds the constraints for the electricity commodity at the storage assets.
        When charging the electricity_storage acts as an electrcicity demand and during discharging
        it acts as a electricity producer. The constraints are selected using the bigM method using
        the boolean for charging and using a charging efficiency during charging.
        """
        constraints = []
        parameters = self.parameters(ensemble_member)

        for asset in [
            *self.energy_system_components.get("electricity_storage", []),
        ]:
            min_voltage = parameters[f"{asset}.min_voltage"]
            voltage = self.state(f"{asset}.ElectricityIn.V")

            # when charging act like a demand, when discharging act like a source
            # to ensure that voltage is equal or larger than the minimum voltage
            constraints.append(((voltage - min_voltage) / min_voltage, 0.0, np.inf))

            power_nom = self.variable_nominal(f"{asset}.ElectricityIn.Power")
            curr_nom = self.variable_nominal(f"{asset}.ElectricityIn.I")
            power_in = self.state(f"{asset}.ElectricityIn.Power")
            current_in = self.state(f"{asset}.ElectricityIn.I")

            # is_charging is 1 if charging and powerin>0
            big_m = 2 * max(np.abs(self.bounds()[f"{asset}.ElectricityIn.Power"]))
            is_charging = self.state(f"{asset}__is_charging")
            constraints.append(((power_in + (1 - is_charging) * big_m) / power_nom, 0.0, np.inf))
            constraints.append(((power_in - is_charging * big_m) / power_nom, -np.inf, 0.0))

            constraints.append(
                (
                    (power_in - min_voltage * current_in + (1 - is_charging) * big_m)
                    / (power_nom * curr_nom * min_voltage) ** 0.5,
                    0,
                    np.inf,
                )
            )
            constraints.append(
                (
                    (power_in - min_voltage * current_in - (1 - is_charging) * big_m)
                    / (power_nom * curr_nom * min_voltage) ** 0.5,
                    -np.inf,
                    0,
                )
            )

            # power charging using discharge/charge efficiency, needs boolean
            eff_power = self.state(f"{asset}.Effective_power_charging")
            discharge_eff = parameters[f"{asset}.discharge_efficiency"]
            charge_eff = parameters[f"{asset}.charge_efficiency"]
            # charging
            constraints.append(
                (
                    (eff_power - charge_eff * power_in + (1 - is_charging) * big_m) / power_nom,
                    0,
                    np.inf,
                )
            )
            constraints.append(
                (
                    (eff_power - charge_eff * power_in - (1 - is_charging) * big_m) / power_nom,
                    -np.inf,
                    0,
                )
            )
            # discharging
            constraints.append(
                (
                    (eff_power * discharge_eff - power_in + is_charging * big_m) / power_nom,
                    0,
                    np.inf,
                )
            )
            constraints.append(
                (
                    (eff_power * discharge_eff - power_in - is_charging * big_m) / power_nom,
                    -np.inf,
                    0,
                )
            )

        return constraints

    def __electricity_storage_discharge_var_path_constraints(
        self, ensemble_member: int
    ) -> List[Tuple[ca.MX, float, float]]:
        """
        The discharge variables are added such that two separate goals for charging and discharging
        can be created. The discharging variable has a lower bound of 0 and should always be larger
        or equal to the negative of the inflow variable. This allows for first a minimization of
        the discharging and afterwards a maximisation of the charging without conflicting goals or
        constraints.

        :param ensemble_member:
        :return: list of the additional constraints that are created
        """

        constraints = []
        options = self.energy_system_options()

        if options["electricity_storage_discharge_variables"]:
            for storage in self.energy_system_components.get("electricity_storage", []):
                storage_eff_power_charge_var = self.state(f"{storage}.Effective_power_charging")
                discharge_var_name = self.__electricity_storage_discharge_map[storage]
                storage_discharge_var = self.__electricity_storage_discharge_var[discharge_var_name]
                nominal = self.variable_nominal(discharge_var_name)

                # P_effective_charge represents both charging and discharing based on the sign.
                # P_discharge >= -P_effective_charge
                constraints.append(
                    ((storage_discharge_var + storage_eff_power_charge_var) / nominal, 0.0, np.inf)
                )

        return constraints

    def __get_electrolyzer_gas_mass_flow_out(
        self, coef_a, coef_b, coef_c, electrical_power_input
    ) -> float:
        """
        This function returns the gas mass flow rate [kg/s] out of an electrolyzer based on the
        theoretical efficiency curve:
        energy [Ws] / gas mass [kg] =
        (coef_a / electrical_power_input) + (b * electrical_power_input) + coef_c

        Parameters
        ----------
        coef_a: electrolyzer efficience curve coefficent
        coef_b: electrolyzer efficience curve coefficent
        coef_c: electrolyzer efficience curve coefficent
        electrical_power_input: electrical power consumed [W]

        Returns
        -------
        gas mass flow rate produced by the electrolyzer [kg/s]
        """

        if not isclose(electrical_power_input, 0.0):
            eff = (
                (coef_a / electrical_power_input) + (coef_b * electrical_power_input) + coef_c
            )  # Wh/g
            gas_mass_flow_out = (1.0 / (eff * 3600)) * electrical_power_input
        else:
            gas_mass_flow_out = 0.0

        return gas_mass_flow_out

    def _get_linear_coef_electrolyzer_mass_vs_epower_fit(
        self, coef_a, coef_b, coef_c, n_lines, electrical_power_min, electrical_power_max
    ) -> Tuple[np.array, np.array]:
        """
        This function returns a set of coefficients to approximate a gas mass flow rate curve with
        linear functions in the form of: gass mass flow rate [kg/s] = b + (a * electrical_power)

        Parameters
        ----------
        coef_a: electrolyzer efficience curve coefficent
        coef_b: electrolyzer efficience curve coefficent
        coef_c: electrolyzer efficience curve coefficent
        n_lines: numebr of linear lines used to approximate the non-linear curve
        electrical_power_min: minimum electrical power consumed [W]
        electrical_power_max: maximum electrical power consumed [W]

        Returns
        -------
        coefficients for linear curve fit(s) to the theoretical non-linear electrolyzer curve
        """

        electrical_power_points = np.linspace(
            electrical_power_min, electrical_power_max, n_lines + 1
        )

        gas_mass_flow_points = np.array(
            [
                self.__get_electrolyzer_gas_mass_flow_out(coef_a, coef_b, coef_c, ep)
                for ep in electrical_power_points
            ]
        )

        a_vals = np.diff(gas_mass_flow_points) / np.diff(electrical_power_points)
        b_vals = gas_mass_flow_points[1:] - a_vals * electrical_power_points[1:]

        return a_vals, b_vals

    def __electrolyzer_path_constaint(self, ensemble_member):
        """
        This functions add the constraints for the gas mass flow production based as a functions of
        electrical power input. This production is approximated by an electrolyzer efficience curve
        (energy/gas mass vs electrical power input, [Ws/kg] vs [W]) which is then linearized. If
        the load becomes lower than the minimum load both the gass_mass_flow and the electricity
        power should be 0.
        """
        constraints = []
        parameters = self.parameters(ensemble_member)
        options = self.energy_system_options()
        # TODO: CHECK UNITS MASSFLOW
        for asset in self.energy_system_components.get("electrolyzer", []):
            gas_mass_flow_out = self.state(f"{asset}.Gas_mass_flow_out")
            power_consumed = self.state(f"{asset}.Power_consumed")
            var_name = self.__asset_is_switched_on_map[asset]
            asset_is_switched_on = self.state(var_name)
            if options["electrolyzer_efficiency"] == ElectrolyzerOption.CONSTANT_EFFICIENCY:
                nominal = (
                    self.variable_nominal(f"{asset}.Gas_mass_flow_out")
                    * self.variable_nominal(f"{asset}.Power_consumed")
                ) ** 0.5 * 3600
                big_m = (
                    self.bounds()[f"{asset}.Power_consumed"][1]
                    / parameters[f"{asset}.efficiency"]
                    / 3600
                ) * 2
                constraints.extend(
                    [
                        (
                            (
                                gas_mass_flow_out * parameters[f"{asset}.efficiency"] * 3600
                                - power_consumed
                            )
                            / nominal,
                            0.0,
                            0.0,
                        ),
                    ]
                )

            elif (
                options["electrolyzer_efficiency"]
                == ElectrolyzerOption.LINEARIZED_THREE_LINES_WEAK_INEQUALITY
            ):
                curve_fit_number_of_lines = 3
                linear_coef_a, linear_coef_b = (
                    self._get_linear_coef_electrolyzer_mass_vs_epower_fit(
                        parameters[f"{asset}.a_eff_coefficient"],
                        parameters[f"{asset}.b_eff_coefficient"],
                        parameters[f"{asset}.c_eff_coefficient"],
                        n_lines=curve_fit_number_of_lines,
                        electrical_power_min=max(
                            parameters[f"{asset}.minimum_load"],
                            0.01 * self.bounds()[f"{asset}.ElectricityIn.Power"][1],
                        ),
                        electrical_power_max=self.bounds()[f"{asset}.ElectricityIn.Power"][1],
                    )
                )
                power_consumed_vect = ca.repmat(power_consumed, len(linear_coef_a))
                gas_mass_flow_out_vect = ca.repmat(gas_mass_flow_out, len(linear_coef_a))
                gass_mass_out_linearized_vect = linear_coef_a * power_consumed_vect + linear_coef_b

                gass_mass_out_max = (
                    linear_coef_a[-1] * self.bounds()[f"{asset}.Power_consumed"][1]
                    + linear_coef_b[-1]
                )
                nominal = (
                    self.variable_nominal(f"{asset}.Gas_mass_flow_out")
                    * min(linear_coef_a)
                    * self.variable_nominal(f"{asset}.Power_consumed")
                ) ** 0.5
                big_m = gass_mass_out_max * 2
                constraints.extend(
                    [
                        (
                            (
                                gas_mass_flow_out_vect
                                - gass_mass_out_linearized_vect
                                - (1 - asset_is_switched_on) * big_m
                            )
                            / nominal,
                            -np.inf,
                            0.0,
                        ),
                    ]
                )
            elif (
                options["electrolyzer_efficiency"]
                == ElectrolyzerOption.LINEARIZED_THREE_LINES_EQUALITY
            ):
                raise NotImplementedError

            constraints.append(
                ((gas_mass_flow_out + asset_is_switched_on * big_m) / big_m, 0.0, np.inf)
            )
            constraints.append(
                ((gas_mass_flow_out - asset_is_switched_on * big_m) / big_m, -np.inf, 0.0)
            )

            # Add constraints to ensure the electrolyzer is switched off when it reaches a power
            # input below the minimum operating value

            big_m = self.bounds()[f"{asset}.ElectricityIn.Power"][1] * 1.5 * 10.0
            constraints.append(
                (
                    (
                        power_consumed
                        - parameters[f"{asset}.minimum_load"]
                        + (1.0 - asset_is_switched_on) * big_m
                    )
                    / self.variable_nominal(f"{asset}.Power_consumed"),
                    0.0,
                    np.inf,
                )
            )
            constraints.append(
                ((power_consumed + asset_is_switched_on * big_m) / big_m, 0.0, np.inf)
            )
            constraints.append(
                ((power_consumed - asset_is_switched_on * big_m) / big_m, -np.inf, 0.0)
            )

        return constraints

    def path_constraints(self, ensemble_member):
        """
        Here we add all the path constraints to the optimization problem. Please note that the
        path constraints are the constraints that are applied to each time-step in the problem.
        """

        constraints = super().path_constraints(ensemble_member)

        constraints.extend(self.__electricity_demand_path_constraints(ensemble_member))
        constraints.extend(self.__electricity_node_mixing_path_constraints(ensemble_member))
        constraints.extend(self.__electricity_cable_mixing_path_constraints(ensemble_member))
        constraints.extend(self.__voltage_loss_path_constraints(ensemble_member))
        constraints.extend(self.__electrolyzer_path_constaint(ensemble_member))
        constraints.extend(self.__electricity_storage_path_constraints(ensemble_member))
        constraints.extend(
            self.__electricity_storage_discharge_var_path_constraints(ensemble_member)
        )

        return constraints

    def constraints(self, ensemble_member):
        """
        This function adds the normal constraints to the problem. Unlike the path constraints these
        are not applied to every time-step in the problem. Meaning that these constraints either
        consider global variables that are independent of time-step or that the relevant time-steps
        are indexed within the constraint formulation.
        """
        constraints = super().constraints(ensemble_member)

        constraints.extend(self.__electricity_producer_set_point_constraints(ensemble_member))

        return constraints

    def goal_programming_options(self):
        """
        Here we set the goal programming configuration. We use soft constraints for consecutive
        goals.
        """
        options = super().goal_programming_options()
        options["keep_soft_constraints"] = True
        return options

    def solver_options(self):
        """
        Here we define the solver options. By default we use the open-source solver cbc and casadi
        solver qpsol.
        """
        options = super().solver_options()
        options["casadi_solver"] = "qpsol"
        options["solver"] = "highs"
        return options

    def compiler_options(self):
        """
        In this function we set the compiler configuration.
        """
        options = super().compiler_options()
        options["resolve_parameter_values"] = True
        return options



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestMultiCommodityHeatPump(TestCase):
    """Test to verify that the optimisation problem can handle multicommodity problems, relating
    electricity and heat"""

    def test_heat_pump_elec_min_heat(self):
        """
        Verify that the minimisation of the heat_source used, and thus the optimization should
        exploit the heatpump as much as possible, and minimum use of heat source at secondary
        side.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - Checks for sufficient production
        - Checks for heat pump energy conservation and COP modelling
        - Checks for Power = I * V at the heatpump

        """
        import models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec as run_hp_elec
        from models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec import HeatProblem2

        base_folder = Path(run_hp_elec.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            HeatProblem2,
            base_folder=base_folder,
            esdl_file_name="heat_pump_elec.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)

        v_min_hp = solution.parameters(0)["GenericConversion_3d3f.min_voltage"]
        i_max = solution.parameters(0)["ElectricityCable_9d3b.max_current"]
        cop = solution.parameters(0)["GenericConversion_3d3f.COP"]

        demand_matching_test(solution, results)
        heat_to_discharge_test(solution, results)
        energy_conservation_test(solution, results)

        heatsource_prim = results["ResidualHeatSource_61b8.Heat_source"]
        heatsource_sec = results["ResidualHeatSource_aec9.Heat_source"]
        heatpump_power = results["GenericConversion_3d3f.Power_elec"]
        heatpump_heat_prim = results["GenericConversion_3d3f.Primary_heat"]
        heatpump_heat_sec = results["GenericConversion_3d3f.Secondary_heat"]
        heatdemand_sec = results["HeatingDemand_18aa.Heat_demand"]
        heatdemand_prim = results["HeatingDemand_3322.Heat_demand"]
        elec_prod_power = results["ElectricityProducer_ac2e.ElectricityOut.Power"]

        heatpump_voltage = results["GenericConversion_3d3f.ElectricityIn.V"]
        heatpump_current = results["GenericConversion_3d3f.ElectricityIn.I"]

        np.testing.assert_allclose(heatsource_sec, 0.0, atol=1.0e-3)

        # check that heatpump is providing more energy to secondary side than demanded
        np.testing.assert_array_less(heatdemand_sec - heatpump_heat_sec, 0)
        # check that producer is providing more energy to heatpump and primary demand
        np.testing.assert_array_less(heatdemand_prim - (heatsource_prim - heatpump_heat_prim), 0)

        # check that heatpumppower*COP==secondaryheat heatpump
        np.testing.assert_allclose(heatpump_power * cop, heatpump_heat_sec)
        # check power consumption with current and voltage heatpump
        np.testing.assert_allclose(heatpump_power, heatpump_current * heatpump_voltage)
        np.testing.assert_array_less(heatpump_power, elec_prod_power)
        # check if current and voltage limits are satisfied
        np.testing.assert_array_less(heatpump_current, i_max * np.ones(len(heatpump_current)))
        np.testing.assert_allclose(v_min_hp * np.ones(len(heatpump_voltage)), heatpump_voltage)

    def test_heat_pump_elec_min_heat_curr_limit(self):
        """
        Verify the minimization of the heat_source used. However, due to limitations in the
        electricity transport through the cables, the power to the heatpump is limited. This in
        turn limits the heat produced by the heatpump which is then not sufficient for the total
        heating demand, resulting in heat production by the secondary heatsource (heat produced by
        this asset is not 0).

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - Checks for sufficient production
        - Checks for heat pump energy conservation and COP modelling
        - Checks for Power = I * V at the heatpump

        """
        import models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec as run_hp_elec
        from models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec import HeatProblem

        base_folder = Path(run_hp_elec.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            HeatProblem,
            base_folder=base_folder,
            esdl_file_name="heat_pump_elec.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)

        v_min_hp = solution.parameters(0)["GenericConversion_3d3f.min_voltage"]
        i_max = solution.parameters(0)["ElectricityCable_9d3b.max_current"]
        cop = solution.parameters(0)["GenericConversion_3d3f.COP"]

        heatsource_prim = results["ResidualHeatSource_61b8.Heat_source"]
        heatsource_sec = results["ResidualHeatSource_aec9.Heat_source"]
        heatpump_power = results["GenericConversion_3d3f.Power_elec"]
        heatpump_heat_prim = results["GenericConversion_3d3f.Primary_heat"]
        heatpump_heat_sec = results["GenericConversion_3d3f.Secondary_heat"]
        heatdemand_sec = results["HeatingDemand_18aa.Heat_demand"]
        heatdemand_prim = results["HeatingDemand_3322.Heat_demand"]
        elec_prod_power = results["ElectricityProducer_ac2e.ElectricityOut.Power"]

        heatpump_voltage = results["GenericConversion_3d3f.ElectricityIn.V"]
        heatpump_current = results["GenericConversion_3d3f.ElectricityIn.I"]

        # check that heatpump isnot providing enough energy to secondary side for demanded
        np.testing.assert_array_less(
            np.zeros(len(heatdemand_sec)), heatdemand_sec - heatpump_heat_sec
        )
        np.testing.assert_array_less(
            heatdemand_sec - (heatpump_heat_sec + heatsource_sec), np.zeros(len(heatdemand_sec))
        )
        # check that heatpump is limited by electric transport power limitations:
        np.testing.assert_allclose(heatpump_power, i_max * v_min_hp * np.ones(len(heatpump_power)))
        # check that prim producer is providing more energy to heatpump and primary demand
        np.testing.assert_array_less(heatdemand_prim - (heatsource_prim - heatpump_heat_prim), 0)
        # check that heatpumppower*COP==secondaryheat heatpump
        np.testing.assert_allclose(heatpump_power * cop, heatpump_heat_sec)
        # check power consumption with current and voltage heatpump
        np.testing.assert_allclose(heatpump_power, heatpump_current * heatpump_voltage)
        np.testing.assert_array_less(heatpump_power, elec_prod_power)
        # check if current and voltage limits are satisfied
        np.testing.assert_allclose(heatpump_current, i_max * np.ones(len(heatpump_current)))
        np.testing.assert_allclose(v_min_hp * np.ones(len(heatpump_voltage)), heatpump_voltage)
        # TODO: currently connecting pipes at HPs can not be disabled, these don't have the
        # functionality as this causes other problems with HP tests, have to adjust this later.
        # This option would be added/changed in asset_to_component_base

    def test_heat_pump_elec_min_elec(self):
        """
        Verify that minimisation of the electricity power used, and thus
        exploiting the heatpump only for heat that cannot directly be covered by other sources.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - Checks for sufficient production
        - Checks for heat pump energy conservation and COP modelling
        - Checks for Power = I * V at the heatpump

        """
        import models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec as run_hp_elec
        from models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec import (
            ElectricityProblem,
        )

        base_folder = Path(run_hp_elec.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            ElectricityProblem,
            base_folder=base_folder,
            esdl_file_name="heat_pump_elec.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)

        tol = 1e-6
        heatsource_prim = results["ResidualHeatSource_61b8.Heat_source"]
        # heatsource_sec = results["ResidualHeatSource_aec9.Heat_source"]
        heatpump_power = results["GenericConversion_3d3f.Power_elec"]
        heatpump_heat_prim = results["GenericConversion_3d3f.Primary_heat"]
        heatpump_heat_sec = results["GenericConversion_3d3f.Secondary_heat"]
        heatpump_disabled = results["GenericConversion_3d3f__disabled"]
        # heatdemand_sec = results["HeatingDemand_18aa.Heat_demand"]
        heatdemand_prim = results["HeatingDemand_3322.Heat_demand"]
        elec_prod_power = results["ElectricityProducer_ac2e.ElectricityOut.Power"]
        # pipe_sec_out_hp_disconnected = results["Pipe_408e__is_disconnected"]

        # check that heatpump is not used:
        np.testing.assert_allclose(heatpump_power, np.zeros(len(heatpump_power)), atol=tol)
        np.testing.assert_allclose(heatpump_heat_sec, np.zeros(len(heatpump_heat_sec)), atol=tol)
        np.testing.assert_allclose(heatpump_heat_prim, np.zeros(len(heatpump_heat_prim)), atol=tol)

        np.testing.assert_allclose(elec_prod_power, np.zeros(len(heatpump_heat_prim)), atol=tol)
        np.testing.assert_allclose(heatpump_disabled, np.ones(len(heatpump_heat_prim)), atol=tol)

        # check that prim producer is providing more energy to heatpump and primary demand
        np.testing.assert_array_less(heatdemand_prim - (heatsource_prim - heatpump_heat_prim), 0)

    def test_heat_pump_elec_price_profile(self):
        """
        Verify that minimisation of the electricity power used when price of electricity is high.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - Check that the producer not connected to the heat pump is only used when electricity
        prices of HP are high
        - Check calculation of variable operational costs which include the electricity prices

        """
        import models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec as run_hp_elec
        from models.unit_cases_electricity.heat_pump_elec.src.run_hp_elec import (
            ElectricityProblemPriceProfile,
        )

        base_folder = Path(run_hp_elec.__file__).resolve().parent.parent

        solution = run_esdl_mesido_optimization(
            ElectricityProblemPriceProfile,
            base_folder=base_folder,
            esdl_file_name="heat_pump_elec_priceprofile.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_elec.csv",
        )

        results = solution.extract_results()

        demand_matching_test(solution, results)
        energy_conservation_test(solution, results)
        heat_to_discharge_test(solution, results)

        heatpump_power = results["GenericConversion_3d3f.Power_elec"]
        heatpump_heat_sec = results["GenericConversion_3d3f.Secondary_heat"]
        heatpump_disabled = results["GenericConversion_3d3f__disabled"]
        heatdemand_sec = results["HeatingDemand_18aa.Heat_demand"]
        var_opex_hp = results["GenericConversion_3d3f__variable_operational_cost"]
        # pipe_sec_out_hp_disconnected = results["Pipe_408e__is_disconnected"]

        # check that heatpump is not used when electricity price is high:
        price_profile = solution.get_timeseries("Electr.price_profile").values
        price_profile_max = price_profile == max(price_profile)
        self.assertTrue(all(price_profile_max >= heatpump_disabled))
        self.assertTrue(all(price_profile_max[1:] * heatpump_power[1:] == 0))

        # check that heatpump is producing all heat for the heatdemand on the secondary side when
        # electricity price is low
        ind_hp = np.asarray(1 - price_profile_max).nonzero()
        np.testing.assert_allclose(heatpump_heat_sec[ind_hp], heatdemand_sec[ind_hp])

        # check variable_operational_cost for heat pump including the price_profile of electricity
        var_opex_hp_non_el = 1e-6  # var_opex in ESDL for HP
        timestep = 1.0  # hr
        var_opex_hp_calc = sum(
            (price_profile[1:] + var_opex_hp_non_el) * heatpump_power[1:] * timestep
        )
        np.testing.assert_allclose(var_opex_hp_calc, var_opex_hp)



from pathlib import Path
from unittest import TestCase

from mesido._heat_loss_u_values_pipe import pipe_heat_loss
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import demand_matching_test, energy_conservation_test, heat_to_discharge_test


class TestVaryingTemperature(TestCase):
    def test_1a_temperature_variation(self):
        """
        This test is to check if the varying network temperature works as expected on a simple
        network. We give it temperature options such that it should select a minimum delta T to be
        able to meet the heat demands. It is known which network temperatures should be selected
        based on the specified input values.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation
        - Check expected supply temperature
        - Check expected return temperature
        - Check on integer variable for selected temperature.
        - Check if the heat losses are correct for the selected temperature

        """
        import models.unit_cases.case_1a.src.run_1a as run_1a
        from models.unit_cases.case_1a.src.run_1a import HeatProblemTvar

        base_folder = Path(run_1a.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblemTvar,
            base_folder=base_folder,
            esdl_file_name="1a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        # This optimization problem is to see whether the correct minimum delta temperaute is
        # chose by the optimization, minimum of 21 deg is needed, only the 85/60 option is
        # feasible, note that a maximum velocity is set low to achieve this.
        results = heat_problem.extract_results()

        test = TestCase()
        test.assertTrue(heat_problem.solver_stats["success"], msg="Optimisation did not succeed")
        # Check that the highest supply temperature is selected
        np.testing.assert_allclose(results[f"{3625334968694477359}_temperature"], 85.0)
        np.testing.assert_allclose(results[f"{3625334968694477359}_85.0"], 1.0)

        # Check that the lowest return temperature is selected
        np.testing.assert_allclose(results[f"{3625334968694477359000}_temperature"], 60.0)
        np.testing.assert_allclose(results[f"{3625334968694477359000}_60.0"], 1.0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        parameters = heat_problem.parameters(0)

        for pipe in heat_problem.energy_system_components.get("heat_pipe", []):
            heat_loss_opt = results[f"{pipe}__hn_heat_loss"]
            carrier_id = parameters[f"{pipe}.carrier_id"]
            temperature = results[f"{carrier_id}_temperature"]
            heat_loss_calc = [
                pipe_heat_loss(
                    heat_problem,
                    heat_problem.energy_system_options(),
                    heat_problem.parameters(0),
                    pipe,
                    None,
                    temp,
                )
                for temp in temperature
            ]
            np.testing.assert_allclose(heat_loss_opt, heat_loss_calc, atol=1.0e-6)

    def test_3a_temperature_variation_supply(self):
        """
        Check varying temperature behoviour for network with storage (tank). In this case we
        Minimise the produced heat and thus we expect the lowest temperature to be selected.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation.
        - Check if the expected temperature is selected and if temperature variable is set
        correctly.
        - Check if the heat losses are correct for the selected temperature

        """
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemTvarsup

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblemTvarsup,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        test = TestCase()
        test.assertTrue(heat_problem.solver_stats["success"], msg="Optimisation did not succeed")

        # optimization with two choices in supply temp 80 and 120 deg
        # lowest temperature should be selected because of lower heat losses and
        # heat production minimization goal
        results = heat_problem.extract_results()

        # Check whehter the heat demand is matched
        for d in heat_problem.energy_system_components.get("heat_demand", []):
            target = heat_problem.get_timeseries(f"{d}.target_heat_demand").values[
                : len(heat_problem.times())
            ]
            np.testing.assert_allclose(target, results[f"{d}.Heat_demand"])

        # Check that the lowest temperature (80.0) is the outputted temperature
        np.testing.assert_allclose(results[f"{4195016129475469474608}_temperature"], 80.0)
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{4195016129475469474608}_80.0"], 1.0)
        np.testing.assert_allclose(results[f"{4195016129475469474608}_120.0"], 0.0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        parameters = heat_problem.parameters(0)

        for pipe in heat_problem.energy_system_components.get("heat_pipe", []):
            heat_loss_opt = results[f"{pipe}__hn_heat_loss"]
            carrier_id = parameters[f"{pipe}.carrier_id"]
            if carrier_id == 4195016129475469474608:
                temperature = results[f"{carrier_id}_temperature"]
                heat_loss_calc = [
                    pipe_heat_loss(
                        heat_problem,
                        heat_problem.energy_system_options(),
                        heat_problem.parameters(0),
                        pipe,
                        None,
                        temp,
                    )
                    for temp in temperature
                ]
                np.testing.assert_allclose(heat_loss_opt, heat_loss_calc, atol=1.0e-6)

    def test_3a_temperature_variation_return(self):
        """
        Check varying temperature behoviour for network with storage (tank). In this case we
        Minimise the flow at the producer, and thus we expect the lowest temperature to be selected
        to maximise the detla T.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation.
        - Check if the expected temperature is selected and if temperature variable is set
        correctly.
        - Check if the heat losses are correct for the selected temperature

        """
        import models.unit_cases.case_3a.src.run_3a as run_3a
        from models.unit_cases.case_3a.src.run_3a import HeatProblemTvarret

        base_folder = Path(run_3a.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblemTvarret,
            base_folder=base_folder,
            esdl_file_name="3a.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        test = TestCase()
        test.assertTrue(heat_problem.solver_stats["success"], msg="Optimisation did not succeed")

        # optimization with two choices in return temp 30 and 40 deg
        # lowest temperature should be selected because of larger dT causing lowest flow rates
        # and we apply source Q minimization goal
        results = heat_problem.extract_results()

        # Check that the lowest temperature (30.0) is the outputted temperature
        np.testing.assert_allclose(results[f"{4195016129475469474608000}_temperature"], 30.0)
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{4195016129475469474608000}_30.0"], 1.0)
        np.testing.assert_allclose(results[f"{4195016129475469474608000}_40.0"], 0.0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        parameters = heat_problem.parameters(0)

        for pipe in heat_problem.energy_system_components.get("heat_pipe", []):
            heat_loss_opt = results[f"{pipe}__hn_heat_loss"]
            carrier_id = parameters[f"{pipe}.carrier_id"]
            if carrier_id == 4195016129475469474608000:
                temperature = results[f"{carrier_id}_temperature"]
                heat_loss_calc = [
                    pipe_heat_loss(
                        heat_problem,
                        heat_problem.energy_system_options(),
                        heat_problem.parameters(0),
                        pipe,
                        None,
                        temp,
                    )
                    for temp in temperature
                ]
                np.testing.assert_allclose(heat_loss_opt, heat_loss_calc, atol=1.0e-6)

    def test_hex_temperature_variation(self):
        """
        This test is to check whether the heat exchanger behaves as expected when optimized under
        varying network temperature. This is of special interest as we want to ensure the
        temperatures stay physically feasible, therefore we create a problem where the lowest
        available supply T is infeasible.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation.
        - Check that the infeasible temperature is not selected.

        """
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import HeatProblemTvar

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblemTvar,
            base_folder=base_folder,
            esdl_file_name="heat_exchanger.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        test = TestCase()
        test.assertTrue(heat_problem.solver_stats["success"], msg="Optimisation did not succeed")

        # optimization with three choices of primary supply temperature of which the lowest is
        # infeasible. Therefore optimization should select second lowest option of 80.
        # lowest feasible temperature should be selected due to heatlosses and the minimization
        # goal of the sources
        results = heat_problem.extract_results()

        # Check that the lowest feasible temperature (80.0) is the outputted temperature and not
        # the 69 which is below the secondary side supply temperature
        np.testing.assert_allclose(results[f"{33638164429859421}_temperature"], 80.0)
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{33638164429859421}_69.0"], 0.0)
        np.testing.assert_allclose(results[f"{33638164429859421}_80.0"], 1.0)
        np.testing.assert_allclose(results[f"{33638164429859421}_90.0"], 0.0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        parameters = heat_problem.parameters(0)

        for pipe in heat_problem.energy_system_components.get("heat_pipe", []):
            heat_loss_opt = results[f"{pipe}__hn_heat_loss"]
            carrier_id = parameters[f"{pipe}.carrier_id"]
            if carrier_id == 33638164429859421:
                temperature = results[f"{carrier_id}_temperature"]
                heat_loss_calc = [
                    pipe_heat_loss(
                        heat_problem,
                        heat_problem.energy_system_options(),
                        heat_problem.parameters(0),
                        pipe,
                        None,
                        temp,
                    )
                    for temp in temperature
                ]
                np.testing.assert_allclose(heat_loss_opt, heat_loss_calc, atol=1.0e-6)

    def test_hex_temperature_variation_disablehex(self):
        """
        This test is to check if the optimizer disables the heat exchanger when only infeasible
        temperature option are provided to it.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation.
        - Infeasible T for heat exchanger is selected.
        - Heat exchanger is disabled.

        """
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import HeatProblemTvarDisableHEX

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblemTvarDisableHEX,
            base_folder=base_folder,
            esdl_file_name="heat_exchanger.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )
        # FIXME: apparantly there is a conflict in the constraints for the is_disabled_hex
        test = TestCase()
        test.assertTrue(heat_problem.solver_stats["success"], msg="Optimisation did not succeed")

        # optimization with only one option in temperature which is infeasible for the hex.
        # therefore optimization should disable the heat exchanger
        results = heat_problem.extract_results()

        # Check that the problem has an infeasible temperature for the hex
        np.testing.assert_allclose(results[f"{33638164429859421}_temperature"], 69.0)
        # Verify that the hex is disabled
        np.testing.assert_allclose(results["HeatExchange_39ed__disabled"], 1.0)
        np.testing.assert_allclose(results["HeatExchange_39ed.Primary_heat"], 0.0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

    def test_hex_temperature_variation_secondary(self):
        """
        Check to see the functioning of the varying network temperature with a heat exchanger where
        all options are feasible and we expect it to take the most advantageous one, which in this
        case is the lowest one.

        Checks:
        - Standard checks for demand matching, heat to discharge and energy conservation.
        - Check that lowest temperature is selected and set correctly.

        """
        import models.heat_exchange.src.run_heat_exchanger as run_heat_exchanger
        from models.heat_exchange.src.run_heat_exchanger import HeatProblemTvarSecondary

        base_folder = Path(run_heat_exchanger.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblemTvarSecondary,
            base_folder=base_folder,
            esdl_file_name="heat_exchanger.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        test = TestCase()
        test.assertTrue(heat_problem.solver_stats["success"], msg="Optimisation did not succeed")

        # optimization with two choices in secondary supply temp 70 and 90 deg
        # lowest temperature should be selected because of heat minimization and lower T has
        # lower heat loss.
        results = heat_problem.extract_results()

        # Check that the lowest temperature (70.0) is the outputted temperature
        np.testing.assert_allclose(results[f"{7212673879469902607010}_temperature"], 70.0)
        # Verify that also the integer is correctly set
        np.testing.assert_allclose(results[f"{7212673879469902607010}_70.0"], 1.0)
        np.testing.assert_allclose(results[f"{7212673879469902607010}_90.0"], 0.0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

    def test_heat_pump_varying_temperature(self):
        """
        Check to see if the Heat pump under varying temperature has the expected COP behaviour.


        """
        import models.heatpump.src.run_heat_pump as run_heat_pump
        from models.heatpump.src.run_heat_pump import HeatProblemTvar

        base_folder = Path(run_heat_pump.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            HeatProblemTvar,
            base_folder=base_folder,
            esdl_file_name="heat_pump.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.xml",
        )

        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        expected_cop = (
            parameters["GenericConversion_3d3f.efficiency"]
            * (273.15 + results[f"{7212673879469902607010}_temperature"])
            / (results[f"{7212673879469902607010}_temperature"] - 70.0)
        )

        np.testing.assert_allclose(
            expected_cop,
            results["GenericConversion_3d3f.Secondary_heat"]
            / results["GenericConversion_3d3f.Power_elec"],
        )

    # Note that CBC struggles heavily and tends to crash, therefore excluded from pipeline
    # def test_varying_temperature_with_pipe_sizing(self):
    #     root_folder = str(Path(__file__).resolve().parent.parent)
    #     import sys
    #     sys.path.insert(1, root_folder)
    #
    #     import examples.pipe_diameter_sizing.src.example  # noqa: E402, I100
    #     from examples.pipe_diameter_sizing.src.example import (
    #         PipeDiameterSizingProblemTvar,
    #     )  # noqa: E402, I100
    #
    #     base_folder = (
    #         Path(examples.pipe_diameter_sizing.src.example.__file__).resolve().parent.parent
    #     )
    #
    #     del root_folder
    #     sys.path.pop(1)
    #
    #     # The optimization should select the highest delta-temperature as
    #     # this minimizes the length
    #     # times diameter cost function
    #
    #     # problem = run_optimization_problem(
    #     PipeDiameterSizingProblemTvar, base_folder=base_folder)
    #     # results = problem.extract_results()
    #
    #
    #     # # Check that the highest supply temperature is selected
    #     # np.testing.assert_allclose(
    #     results[f"{761602374459208051248}__supply_temperature"], 90.0)
    #     #
    #     # # Check that the lowest return temperature is selected
    #     # np.testing.assert_allclose(
    #     results[f"{761602374459208051248000}__return_temperature"], 30.0)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestVaryingTemperature()
    a.test_1a_temperature_variation()
    a.test_3a_temperature_variation_supply()
    a.test_3a_temperature_variation_return()
    a.test_hex_temperature_variation()
    a.test_hex_temperature_variation_disablehex()
    a.test_hex_temperature_variation_secondary()
    a.test_heat_pump_varying_temperature()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from .air_water_heat_pump import AirWaterHeatPump
from .airco import Airco
from .ates import ATES
from .check_valve import CheckValve
from .cold_demand import ColdDemand
from .control_valve import ControlValve
from .geothermal_source import GeothermalSource
from .heat_buffer import HeatBuffer
from .heat_demand import HeatDemand
from .heat_exchanger import HeatExchanger
from .heat_four_port import HeatFourPort
from .heat_pipe import HeatPipe
from .heat_port import HeatPort
from .heat_pump import HeatPump
from .heat_source import HeatSource
from .heat_two_port import HeatTwoPort
from .low_temperature_ates import LowTemperatureATES
from .node import Node
from .pump import Pump

__all__ = [
    "Airco",
    "AirWaterHeatPump",
    "ATES",
    "HeatBuffer",
    "CheckValve",
    "ColdDemand",
    "ControlValve",
    "HeatDemand",
    "GeothermalSource",
    "HeatExchanger",
    "HeatFourPort",
    "HeatPort",
    "HeatPump",
    "HeatTwoPort",
    "HeatPipe",
    "HeatSource",
    "LowTemperatureATES",
    "Node",
    "Pump",
]



import datetime
import unittest
from pathlib import Path
from typing import Optional

import esdl

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import InfluxDBProfileReader, ProfileReaderFromFile
from mesido.workflows import EndScenarioSizingStagedHIGHS

import numpy as np

import pandas as pd


class MockInfluxDBProfileReader(InfluxDBProfileReader):
    def __init__(self, energy_system: esdl.EnergySystem, file_path: Optional[Path]):
        super().__init__(energy_system, file_path)
        self._loaded_profiles = pd.read_csv(
            file_path,
            index_col="DateTime",
            date_parser=lambda x: pd.to_datetime(x).tz_convert(datetime.timezone.utc),
        )

    def _load_profile_timeseries_from_database(self, profile: esdl.InfluxDBProfile) -> pd.Series:
        return self._loaded_profiles[profile.id]


class TestProfileLoading(unittest.TestCase):
    def test_loading_from_influx(self):
        """
        This test checks if loading an ESDL with influxDB profiles works. Since
        the test environment doesn't always have access to an influxDB server,
        a connection is mocked and profiles are instead loaded from the file
        "influx_mock.csv" in the models/unit_cases/case_1a/input folder.
        EndScenarioSizing is called thus the checks done are on profile lenghts
        and some values. This is because this scenario should also do aggragation
        of the profiles for non-peak days.

        Also check:
            - that the timezone setting from the "influx_mock.csv" is correct
        """
        import models.unit_cases.case_1a.src.run_1a as run_1a

        base_folder = Path(run_1a.__file__).resolve().parent.parent
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"
        problem = EndScenarioSizingStagedHIGHS(
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
            esdl_file_name="1a_with_influx_profiles.esdl",
            profile_reader=MockInfluxDBProfileReader,
            input_timeseries_file="influx_mock.csv",
        )
        problem.pre()

        np.testing.assert_equal(problem.io.reference_datetime.tzinfo, datetime.timezone.utc)

        # the three demands in the test ESDL
        for demand_name in ["HeatingDemand_2ab9", "HeatingDemand_6662", "HeatingDemand_506c"]:
            profile_values = problem.get_timeseries(f"{demand_name}.target_heat_demand").values
            self.assertEqual(profile_values[0], 0.0)
            self.assertEqual(len(profile_values), 26)

        heat_price_profile = problem.get_timeseries("Heat.price_profile").values
        self.assertEqual(heat_price_profile[0], 0.0)
        self.assertLess(max(heat_price_profile), 1.0)

    def test_loading_from_csv(self):
        """
        This test constructs a problem with input profiles read from a CSV file.
        The test checks if the profiles read match the profiles from the CVS file and that the
        default UTC timezone has been set.
        """
        import models.unit_cases_electricity.electrolyzer.src.example as example
        from models.unit_cases_electricity.electrolyzer.src.example import MILPProblemInequality

        base_folder = Path(example.__file__).resolve().parent.parent
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"
        problem = MILPProblemInequality(
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
            esdl_file_name="h2.esdl",
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        problem.pre()

        np.testing.assert_equal(problem.io.reference_datetime.tzinfo, datetime.timezone.utc)

        expected_array = np.array([1.0e8] * 3)
        np.testing.assert_equal(
            expected_array,
            problem.get_timeseries("WindPark_7f14.maximum_electricity_source").values,
        )

        expected_array = np.array([1.0] * 3)
        np.testing.assert_equal(expected_array, problem.get_timeseries("elec.price_profile").values)

        expected_array = np.array([1.0e6] * 3)
        np.testing.assert_equal(
            expected_array, problem.get_timeseries("Hydrogen.price_profile").values
        )

    def test_loading_from_xml(self):
        """
        This test loads a simple problem using an XML file for input profiles.
        The test checks if the load profiles match those specified in the XML file and that the
        default UTC timezone has been set.
        """
        import models.basic_source_and_demand.src.heat_comparison as heat_comparison
        from models.basic_source_and_demand.src.heat_comparison import HeatESDL

        base_folder = Path(heat_comparison.__file__).resolve().parent.parent
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"
        problem = HeatESDL(
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
            esdl_file_name="model.esdl",
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.xml",
        )
        problem.pre()

        np.testing.assert_equal(problem.io.reference_datetime.tzinfo, datetime.timezone.utc)

        expected_array = np.array([1.5e5] * 16 + [1.0e5] * 13 + [0.5e5] * 16)
        np.testing.assert_equal(
            expected_array, problem.get_timeseries("demand.target_heat_demand").values
        )

    def test_loading_from_csv_with_influx_profiles_given(self):
        """
        This test loads a problem using an ESDL file which has influxDB profiles
        specified. Furthermore, the problem is given a csv file to load profiles
        from. This test thus checks if the ESDL_mixin correctly loads the profiles
        from the csv instead of trying to get them from influxDB. The test check
        if the loaded profiles match those specified in the csv.
        """
        import models.unit_cases_electricity.electrolyzer.src.example as example
        from models.unit_cases_electricity.electrolyzer.src.example import MILPProblemInequality

        base_folder = Path(example.__file__).resolve().parent.parent
        model_folder = base_folder / "model"
        input_folder = base_folder / "input"
        problem = MILPProblemInequality(
            esdl_parser=ESDLFileParser,
            base_folder=base_folder,
            model_folder=model_folder,
            input_folder=input_folder,
            esdl_file_name="h2_profiles_added_dummy_values.esdl",
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )
        problem.pre()

        np.testing.assert_equal(problem.io.reference_datetime.tzinfo, datetime.timezone.utc)

        expected_array = np.array([1.0e8] * 3)
        np.testing.assert_equal(
            expected_array,
            problem.get_timeseries("WindPark_7f14.maximum_electricity_source").values,
        )

        expected_array = np.array([1.0] * 3)
        np.testing.assert_equal(expected_array, problem.get_timeseries("elec.price_profile").values)

        expected_array = np.array([1.0e6] * 3)
        np.testing.assert_equal(
            expected_array, problem.get_timeseries("Hydrogen.price_profile").values
        )


if __name__ == "__main__":
    # unittest.main()
    a = TestProfileLoading()
    a.test_loading_from_influx()
    a.test_loading_from_csv()
    a.test_loading_from_xml()
    a.test_loading_from_csv_with_influx_profiles_given()



from mesido.pycml import Variable

from numpy import nan

from .electricity_base import ElectricityPort
from .._internal import BaseAsset
from .._internal.electricity_component import ElectricityComponent


class ElectricityStorage(ElectricityComponent, BaseAsset):
    """
    The electricity storage component is used to store electrical power of a network.
    The change in stored electrical power should be equal to the electricity entering and leaving
    the component multiplied with its efficiency
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_storage"

        self.max_capacity = nan
        self.min_voltage = nan
        self.charge_efficiency = 1.0
        self.discharge_efficiency = 1.0

        self.add_variable(ElectricityPort, "ElectricityIn")

        self._typical_fill_time = 3600.0
        self._nominal_stored_electricity = self.ElectricityIn.Power.max * self._typical_fill_time
        self.add_variable(
            Variable,
            "Stored_electricity",
            min=0.0,
            max=self.max_capacity,
            nominal=self._nominal_stored_electricity,
        )
        self.add_variable(
            Variable,
            "Effective_power_charging",
            nominal=self.ElectricityIn.Power.nominal,
            max=self.ElectricityIn.Power.max,
        )

        self.add_equation(
            (
                (self.der(self.Stored_electricity) - self.Effective_power_charging)
                / self.ElectricityIn.Power.nominal
            )
        )



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np


class TestAssetIsRealized(TestCase):
    def test_asset_is_realized(self):
        r"""
        This is a test to check the behaviour of the cumulative investments made and the
        asset is realized variable. We want the asset only to become available once sufficient
        investments are made.

        In this specific test we optimize to match the heat demand. However, the sources are not
        available from the start as the cumulative invesments made at timestep 0 is 0. Furthermore,
        there is a cap on the investments that can be done per timestep. We expect the optimizer
        to find a solution that releases the sources as soon as possible in order to match demand
        and the demand not to be matched until that point in time.

        Checks:
        - That the investment cap per time step is satisfied
        - Check that the total investments actually are sufficient once the asset is realized
        - Check that the asset is not used before the asset_is_realized == 1
        """
        import models.test_case_small_network_with_ates.src.run_ates as run_ates
        from models.test_case_small_network_with_ates.src.run_ates import (
            HeatProblemPlacingOverTime,
        )

        base_folder = Path(run_ates.__file__).resolve().parent.parent

        # This is an optimization done over 25 timesteps with a cap on how quickly the cost
        # for the 2 producers can be realized
        solution = run_esdl_mesido_optimization(
            HeatProblemPlacingOverTime,
            base_folder=base_folder,
            esdl_file_name="test_case_small_network_with_ates.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="Warmte_test.csv",
        )

        results = solution.extract_results()

        # First we test whether the investments made are below cap
        cap = 2.5e5 + 1.0e-3  # some small tolerance, CBC...
        np.testing.assert_allclose(
            True, np.diff(results["HeatProducer_1__cumulative_investments_made_in_eur"]) <= cap
        )
        np.testing.assert_allclose(
            True, np.diff(results["HeatProducer_2__cumulative_investments_made_in_eur"]) <= cap
        )

        # Now we test if the investments made are greater then the needed investments once the
        # asset is realized
        inds_1 = np.where(np.round(results["HeatProducer_1__asset_is_realized"]) == 1)
        np.testing.assert_allclose(
            True,
            (
                results["HeatProducer_1__cumulative_investments_made_in_eur"][inds_1]
                >= results["HeatProducer_1__investment_cost"]
                + results["HeatProducer_1__installation_cost"]
                - 1.0e-3
            ),
        )
        inds_2 = np.where(np.round(results["HeatProducer_2__asset_is_realized"]) == 1)
        np.testing.assert_allclose(
            True,
            (
                results["HeatProducer_2__cumulative_investments_made_in_eur"][inds_2]
                >= results["HeatProducer_2__investment_cost"]
                + results["HeatProducer_2__installation_cost"]
                - 1.0e-3
            ),
        )

        # Here we test that the asset is not used until it is actually realized
        inds_not_1 = np.where(np.round(results["HeatProducer_1__asset_is_realized"]) == 0)
        np.testing.assert_allclose(
            results["HeatProducer_1.Heat_source"][inds_not_1], 0.0, atol=1e-6
        )
        inds_not_2 = np.where(np.round(results["HeatProducer_2__asset_is_realized"]) == 0)
        np.testing.assert_allclose(
            results["HeatProducer_1.Heat_source"][inds_not_2], 0.0, atol=1e-6
        )

        # Here we test that the asset is actually used once it is realized
        np.testing.assert_allclose(results["HeatProducer_1.Heat_source"][inds_1] > 0.0, True)
        np.testing.assert_allclose(results["HeatProducer_2.Heat_source"][inds_2] > 0.0, True)


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestAssetIsRealized()
    a.test_asset_is_realized()
    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))
